@@ -1,9 +1,9 @@
 // Take a look at the license at the top of the repository in the LICENSE file.
 
-use std::fmt;
-use std::ptr;
+use std::{fmt, ptr};
 
-use glib::translate::{from_glib, from_glib_none, FromGlib, IntoGlib, ToGlibPtr};
+use crate::ffi;
+use glib::translate::*;
 use gst::prelude::*;
 
 #[repr(transparent)]
@@ -113,48 +113,57 @@ impl VideoMeta {
     }
 
     #[doc(alias = "get_flags")]
+    #[inline]
     pub fn video_frame_flags(&self) -> crate::VideoFrameFlags {
         unsafe { from_glib(self.0.flags) }
     }
 
     #[doc(alias = "get_format")]
+    #[inline]
     pub fn format(&self) -> crate::VideoFormat {
         unsafe { from_glib(self.0.format) }
     }
 
     #[doc(alias = "get_id")]
+    #[inline]
     pub fn id(&self) -> i32 {
         self.0.id
     }
 
     #[doc(alias = "get_width")]
+    #[inline]
     pub fn width(&self) -> u32 {
         self.0.width
     }
 
     #[doc(alias = "get_height")]
+    #[inline]
     pub fn height(&self) -> u32 {
         self.0.height
     }
 
     #[doc(alias = "get_n_planes")]
+    #[inline]
     pub fn n_planes(&self) -> u32 {
         self.0.n_planes
     }
 
     #[doc(alias = "get_offset")]
+    #[inline]
     pub fn offset(&self) -> &[usize] {
         &self.0.offset[0..(self.0.n_planes as usize)]
     }
 
     #[doc(alias = "get_stride")]
+    #[inline]
     pub fn stride(&self) -> &[i32] {
         &self.0.stride[0..(self.0.n_planes as usize)]
     }
 
-    #[cfg(any(feature = "v1_18", feature = "dox"))]
-    #[cfg_attr(feature = "dox", doc(cfg(feature = "v1_18")))]
+    #[cfg(feature = "v1_18")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "v1_18")))]
     #[doc(alias = "get_alignment")]
+    #[inline]
     pub fn alignment(&self) -> crate::VideoAlignment {
         crate::VideoAlignment::new(
             self.0.alignment.padding_top,
@@ -165,8 +174,8 @@ impl VideoMeta {
         )
     }
 
-    #[cfg(any(feature = "v1_18", feature = "dox"))]
-    #[cfg_attr(feature = "dox", doc(cfg(feature = "v1_18")))]
+    #[cfg(feature = "v1_18")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "v1_18")))]
     #[doc(alias = "get_plane_size")]
     #[doc(alias = "gst_video_meta_get_plane_size")]
     pub fn plane_size(&self) -> Result<[usize; crate::VIDEO_MAX_PLANES], glib::BoolError> {
@@ -174,10 +183,7 @@ impl VideoMeta {
 
         unsafe {
             glib::result_from_gboolean!(
-                ffi::gst_video_meta_get_plane_size(
-                    &self.0 as *const _ as usize as *mut _,
-                    &mut plane_size,
-                ),
+                ffi::gst_video_meta_get_plane_size(mut_override(&self.0), &mut plane_size,),
                 "Failed to get plane size"
             )?;
         }
@@ -185,8 +191,8 @@ impl VideoMeta {
         Ok(plane_size)
     }
 
-    #[cfg(any(feature = "v1_18", feature = "dox"))]
-    #[cfg_attr(feature = "dox", doc(cfg(feature = "v1_18")))]
+    #[cfg(feature = "v1_18")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "v1_18")))]
     #[doc(alias = "get_plane_height")]
     #[doc(alias = "gst_video_meta_get_plane_height")]
     pub fn plane_height(&self) -> Result<[u32; crate::VIDEO_MAX_PLANES], glib::BoolError> {
@@ -194,10 +200,7 @@ impl VideoMeta {
 
         unsafe {
             glib::result_from_gboolean!(
-                ffi::gst_video_meta_get_plane_height(
-                    &self.0 as *const _ as usize as *mut _,
-                    &mut plane_height,
-                ),
+                ffi::gst_video_meta_get_plane_height(mut_override(&self.0), &mut plane_height,),
                 "Failed to get plane height"
             )?;
         }
@@ -205,8 +208,8 @@ impl VideoMeta {
         Ok(plane_height)
     }
 
-    #[cfg(any(feature = "v1_18", feature = "dox"))]
-    #[cfg_attr(feature = "dox", doc(cfg(feature = "v1_18")))]
+    #[cfg(feature = "v1_18")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "v1_18")))]
     #[doc(alias = "gst_video_meta_set_alignment")]
     pub fn set_alignment(
         &mut self,
@@ -225,6 +228,7 @@ unsafe impl MetaAPI for VideoMeta {
     type GstType = ffi::GstVideoMeta;
 
     #[doc(alias = "gst_video_meta_api_get_type")]
+    #[inline]
     fn meta_api() -> glib::Type {
         unsafe { from_glib(ffi::gst_video_meta_api_get_type()) }
     }
@@ -279,10 +283,12 @@ impl VideoCropMeta {
     }
 
     #[doc(alias = "get_rect")]
+    #[inline]
     pub fn rect(&self) -> (u32, u32, u32, u32) {
         (self.0.x, self.0.y, self.0.width, self.0.height)
     }
 
+    #[inline]
     pub fn set_rect(&mut self, rect: (u32, u32, u32, u32)) {
         self.0.x = rect.0;
         self.0.y = rect.1;
@@ -295,6 +301,7 @@ unsafe impl MetaAPI for VideoCropMeta {
     type GstType = ffi::GstVideoCropMeta;
 
     #[doc(alias = "gst_video_crop_meta_api_get_type")]
+    #[inline]
     fn meta_api() -> glib::Type {
         unsafe { from_glib(ffi::gst_video_crop_meta_api_get_type()) }
     }
@@ -338,40 +345,44 @@ impl VideoRegionOfInterestMeta {
     }
 
     #[doc(alias = "get_rect")]
+    #[inline]
     pub fn rect(&self) -> (u32, u32, u32, u32) {
         (self.0.x, self.0.y, self.0.w, self.0.h)
     }
 
     #[doc(alias = "get_id")]
+    #[inline]
     pub fn id(&self) -> i32 {
         self.0.id
     }
 
     #[doc(alias = "get_parent_id")]
+    #[inline]
     pub fn parent_id(&self) -> i32 {
         self.0.parent_id
     }
 
     #[doc(alias = "get_roi_type")]
+    #[inline]
     pub fn roi_type<'a>(&self) -> &'a str {
         unsafe { glib::Quark::from_glib(self.0.roi_type).as_str() }
     }
 
-    #[cfg(feature = "v1_14")]
     #[doc(alias = "get_params")]
     pub fn params(&self) -> ParamsIter {
         ParamsIter {
             _meta: self,
-            list: self.0.params,
+            list: ptr::NonNull::new(self.0.params),
         }
     }
 
-    #[cfg(feature = "v1_14")]
     #[doc(alias = "get_param")]
-    pub fn param<'b>(&self, name: &'b str) -> Option<&gst::StructureRef> {
+    #[inline]
+    pub fn param<'b>(&'b self, name: &str) -> Option<&'b gst::StructureRef> {
         self.params().find(|s| s.name() == name)
     }
 
+    #[inline]
     pub fn set_rect(&mut self, rect: (u32, u32, u32, u32)) {
         self.0.x = rect.0;
         self.0.y = rect.1;
@@ -379,54 +390,54 @@ impl VideoRegionOfInterestMeta {
         self.0.h = rect.3;
     }
 
+    #[inline]
     pub fn set_id(&mut self, id: i32) {
         self.0.id = id
     }
 
+    #[inline]
     pub fn set_parent_id(&mut self, id: i32) {
         self.0.parent_id = id
     }
 
-    #[cfg(feature = "v1_14")]
     #[doc(alias = "gst_video_region_of_interest_meta_add_param")]
     pub fn add_param(&mut self, s: gst::Structure) {
         unsafe {
-            ffi::gst_video_region_of_interest_meta_add_param(&mut self.0, s.into_ptr());
+            ffi::gst_video_region_of_interest_meta_add_param(&mut self.0, s.into_glib_ptr());
         }
     }
 }
 
-#[cfg(feature = "v1_14")]
 pub struct ParamsIter<'a> {
     _meta: &'a VideoRegionOfInterestMeta,
-    list: *const glib::ffi::GList,
+    list: Option<ptr::NonNull<glib::ffi::GList>>,
 }
 
-#[cfg(feature = "v1_14")]
 impl<'a> Iterator for ParamsIter<'a> {
     type Item = &'a gst::StructureRef;
 
     fn next(&mut self) -> Option<&'a gst::StructureRef> {
-        if self.list.is_null() {
-            return None;
-        }
+        match self.list {
+            None => None,
+            Some(list) => unsafe {
+                self.list = ptr::NonNull::new(list.as_ref().next);
+                let data = list.as_ref().data;
 
-        unsafe {
-            let data = (*self.list).data;
-            assert!(!data.is_null());
-            self.list = (*self.list).next;
+                let s = gst::StructureRef::from_glib_borrow(data as *const gst::ffi::GstStructure);
 
-            let s = gst::StructureRef::from_glib_borrow(data as *const gst::ffi::GstStructure);
-
-            Some(s)
+                Some(s)
+            },
         }
     }
 }
 
+impl std::iter::FusedIterator for ParamsIter<'_> {}
+
 unsafe impl MetaAPI for VideoRegionOfInterestMeta {
     type GstType = ffi::GstVideoRegionOfInterestMeta;
 
     #[doc(alias = "gst_video_region_of_interest_meta_api_get_type")]
+    #[inline]
     fn meta_api() -> glib::Type {
         unsafe { from_glib(ffi::gst_video_region_of_interest_meta_api_get_type()) }
     }
@@ -434,19 +445,13 @@ unsafe impl MetaAPI for VideoRegionOfInterestMeta {
 
 impl fmt::Debug for VideoRegionOfInterestMeta {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        let mut f = f.debug_struct("VideoRegionOfInterestMeta");
-
-        f.field("roi_type", &self.roi_type())
+        f.debug_struct("VideoRegionOfInterestMeta")
+            .field("roi_type", &self.roi_type())
             .field("rect", &self.rect())
             .field("id", &self.id())
-            .field("parent_id", &self.parent_id());
-
-        #[cfg(feature = "v1_14")]
-        {
-            f.field("params", &self.params().collect::<Vec<_>>());
-        }
-
-        f.finish()
+            .field("parent_id", &self.parent_id())
+            .field("params", &self.params().collect::<Vec<_>>())
+            .finish()
     }
 }
 
@@ -483,10 +488,12 @@ impl VideoAffineTransformationMeta {
     }
 
     #[doc(alias = "get_matrix")]
+    #[inline]
     pub fn matrix(&self) -> &[[f32; 4]; 4] {
         unsafe { &*(&self.0.matrix as *const [f32; 16] as *const [[f32; 4]; 4]) }
     }
 
+    #[inline]
     pub fn set_matrix(&mut self, matrix: &[[f32; 4]; 4]) {
         for (i, o) in Iterator::zip(matrix.iter().flatten(), self.0.matrix.iter_mut()) {
             *o = *i;
@@ -508,6 +515,7 @@ unsafe impl MetaAPI for VideoAffineTransformationMeta {
     type GstType = ffi::GstVideoAffineTransformationMeta;
 
     #[doc(alias = "gst_video_affine_transformation_meta_api_get_type")]
+    #[inline]
     fn meta_api() -> glib::Type {
         unsafe { from_glib(ffi::gst_video_affine_transformation_meta_api_get_type()) }
     }
@@ -546,15 +554,18 @@ impl VideoOverlayCompositionMeta {
     }
 
     #[doc(alias = "get_overlay")]
+    #[inline]
     pub fn overlay(&self) -> &crate::VideoOverlayCompositionRef {
         unsafe { crate::VideoOverlayCompositionRef::from_ptr(self.0.overlay) }
     }
 
     #[doc(alias = "get_overlay_owned")]
+    #[inline]
     pub fn overlay_owned(&self) -> crate::VideoOverlayComposition {
         unsafe { from_glib_none(self.overlay().as_ptr()) }
     }
 
+    #[inline]
     pub fn set_overlay(&mut self, overlay: &crate::VideoOverlayComposition) {
         #![allow(clippy::cast_ptr_alignment)]
         unsafe {
@@ -569,6 +580,7 @@ unsafe impl MetaAPI for VideoOverlayCompositionMeta {
     type GstType = ffi::GstVideoOverlayCompositionMeta;
 
     #[doc(alias = "gst_video_overlay_composition_meta_api_get_type")]
+    #[inline]
     fn meta_api() -> glib::Type {
         unsafe { from_glib(ffi::gst_video_overlay_composition_meta_api_get_type()) }
     }
@@ -582,21 +594,21 @@ impl fmt::Debug for VideoOverlayCompositionMeta {
     }
 }
 
-#[cfg(any(feature = "v1_16", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_16")))]
+#[cfg(feature = "v1_16")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_16")))]
 #[repr(transparent)]
 #[doc(alias = "GstVideoCaptionMeta")]
 pub struct VideoCaptionMeta(ffi::GstVideoCaptionMeta);
 
-#[cfg(any(feature = "v1_16", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_16")))]
+#[cfg(feature = "v1_16")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_16")))]
 unsafe impl Send for VideoCaptionMeta {}
-#[cfg(any(feature = "v1_16", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_16")))]
+#[cfg(feature = "v1_16")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_16")))]
 unsafe impl Sync for VideoCaptionMeta {}
 
-#[cfg(any(feature = "v1_16", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_16")))]
+#[cfg(feature = "v1_16")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_16")))]
 impl VideoCaptionMeta {
     #[doc(alias = "gst_buffer_add_video_caption_meta")]
     pub fn add<'a>(
@@ -619,12 +631,17 @@ impl VideoCaptionMeta {
     }
 
     #[doc(alias = "get_caption_type")]
+    #[inline]
     pub fn caption_type(&self) -> crate::VideoCaptionType {
         unsafe { from_glib(self.0.caption_type) }
     }
 
     #[doc(alias = "get_data")]
+    #[inline]
     pub fn data(&self) -> &[u8] {
+        if self.0.size == 0 {
+            return &[];
+        }
         unsafe {
             use std::slice;
 
@@ -633,19 +650,20 @@ impl VideoCaptionMeta {
     }
 }
 
-#[cfg(any(feature = "v1_16", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_16")))]
+#[cfg(feature = "v1_16")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_16")))]
 unsafe impl MetaAPI for VideoCaptionMeta {
     type GstType = ffi::GstVideoCaptionMeta;
 
     #[doc(alias = "gst_video_caption_meta_api_get_type")]
+    #[inline]
     fn meta_api() -> glib::Type {
         unsafe { from_glib(ffi::gst_video_caption_meta_api_get_type()) }
     }
 }
 
-#[cfg(any(feature = "v1_16", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_16")))]
+#[cfg(feature = "v1_16")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_16")))]
 impl fmt::Debug for VideoCaptionMeta {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         f.debug_struct("VideoCaptionMeta")
@@ -655,21 +673,21 @@ impl fmt::Debug for VideoCaptionMeta {
     }
 }
 
-#[cfg(any(feature = "v1_18", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_18")))]
+#[cfg(feature = "v1_18")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_18")))]
 #[repr(transparent)]
 #[doc(alias = "GstVideoAFDMeta")]
 pub struct VideoAFDMeta(ffi::GstVideoAFDMeta);
 
-#[cfg(any(feature = "v1_18", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_18")))]
+#[cfg(feature = "v1_18")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_18")))]
 unsafe impl Send for VideoAFDMeta {}
-#[cfg(any(feature = "v1_18", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_18")))]
+#[cfg(feature = "v1_18")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_18")))]
 unsafe impl Sync for VideoAFDMeta {}
 
-#[cfg(any(feature = "v1_18", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_18")))]
+#[cfg(feature = "v1_18")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_18")))]
 impl VideoAFDMeta {
     #[doc(alias = "gst_buffer_add_video_afd_meta")]
     pub fn add(
@@ -693,34 +711,38 @@ impl VideoAFDMeta {
     }
 
     #[doc(alias = "get_field")]
+    #[inline]
     pub fn field(&self) -> u8 {
         self.0.field
     }
 
     #[doc(alias = "get_spec")]
+    #[inline]
     pub fn spec(&self) -> crate::VideoAFDSpec {
         unsafe { from_glib(self.0.spec) }
     }
 
     #[doc(alias = "get_afd")]
+    #[inline]
     pub fn afd(&self) -> crate::VideoAFDValue {
         unsafe { from_glib(self.0.afd) }
     }
 }
 
-#[cfg(any(feature = "v1_18", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_18")))]
+#[cfg(feature = "v1_18")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_18")))]
 unsafe impl MetaAPI for VideoAFDMeta {
     type GstType = ffi::GstVideoAFDMeta;
 
     #[doc(alias = "gst_video_afd_meta_api_get_type")]
+    #[inline]
     fn meta_api() -> glib::Type {
         unsafe { from_glib(ffi::gst_video_afd_meta_api_get_type()) }
     }
 }
 
-#[cfg(any(feature = "v1_18", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_18")))]
+#[cfg(feature = "v1_18")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_18")))]
 impl fmt::Debug for VideoAFDMeta {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         f.debug_struct("VideoAFDMeta")
@@ -731,21 +753,21 @@ impl fmt::Debug for VideoAFDMeta {
     }
 }
 
-#[cfg(any(feature = "v1_18", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_18")))]
+#[cfg(feature = "v1_18")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_18")))]
 #[repr(transparent)]
 #[doc(alias = "GstVideoBarMeta")]
 pub struct VideoBarMeta(ffi::GstVideoBarMeta);
 
-#[cfg(any(feature = "v1_18", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_18")))]
+#[cfg(feature = "v1_18")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_18")))]
 unsafe impl Send for VideoBarMeta {}
-#[cfg(any(feature = "v1_18", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_18")))]
+#[cfg(feature = "v1_18")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_18")))]
 unsafe impl Sync for VideoBarMeta {}
 
-#[cfg(any(feature = "v1_18", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_18")))]
+#[cfg(feature = "v1_18")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_18")))]
 impl VideoBarMeta {
     #[doc(alias = "gst_buffer_add_video_bar_meta")]
     pub fn add(
@@ -771,38 +793,43 @@ impl VideoBarMeta {
     }
 
     #[doc(alias = "get_field")]
+    #[inline]
     pub fn field(&self) -> u8 {
         self.0.field
     }
 
+    #[inline]
     pub fn is_letterbox(&self) -> bool {
         unsafe { from_glib(self.0.is_letterbox) }
     }
 
     #[doc(alias = "get_bar_data1")]
+    #[inline]
     pub fn bar_data1(&self) -> u32 {
         self.0.bar_data1
     }
 
     #[doc(alias = "get_bar_data2")]
+    #[inline]
     pub fn bar_data2(&self) -> u32 {
         self.0.bar_data2
     }
 }
 
-#[cfg(any(feature = "v1_18", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_18")))]
+#[cfg(feature = "v1_18")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_18")))]
 unsafe impl MetaAPI for VideoBarMeta {
     type GstType = ffi::GstVideoBarMeta;
 
     #[doc(alias = "gst_video_bar_meta_api_get_type")]
+    #[inline]
     fn meta_api() -> glib::Type {
         unsafe { from_glib(ffi::gst_video_bar_meta_api_get_type()) }
     }
 }
 
-#[cfg(any(feature = "v1_18", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_18")))]
+#[cfg(feature = "v1_18")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_18")))]
 impl fmt::Debug for VideoBarMeta {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         f.debug_struct("VideoBarMeta")
@@ -814,21 +841,21 @@ impl fmt::Debug for VideoBarMeta {
     }
 }
 
-#[cfg(any(feature = "v1_20", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_20")))]
+#[cfg(feature = "v1_20")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_20")))]
 #[repr(transparent)]
 #[doc(alias = "GstVideoCodecAlphaMeta")]
 pub struct VideoCodecAlphaMeta(ffi::GstVideoCodecAlphaMeta);
 
-#[cfg(any(feature = "v1_20", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_20")))]
+#[cfg(feature = "v1_20")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_20")))]
 unsafe impl Send for VideoCodecAlphaMeta {}
-#[cfg(any(feature = "v1_20", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_20")))]
+#[cfg(feature = "v1_20")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_20")))]
 unsafe impl Sync for VideoCodecAlphaMeta {}
 
-#[cfg(any(feature = "v1_20", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_20")))]
+#[cfg(feature = "v1_20")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_20")))]
 impl VideoCodecAlphaMeta {
     #[doc(alias = "gst_buffer_add_video_codec_alpha_meta")]
     pub fn add(
@@ -846,28 +873,31 @@ impl VideoCodecAlphaMeta {
         }
     }
 
+    #[inline]
     pub fn alpha_buffer(&self) -> &gst::BufferRef {
         unsafe { gst::BufferRef::from_ptr(self.0.buffer) }
     }
 
+    #[inline]
     pub fn alpha_buffer_owned(&self) -> gst::Buffer {
         unsafe { from_glib_none(self.0.buffer) }
     }
 }
 
-#[cfg(any(feature = "v1_20", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_20")))]
+#[cfg(feature = "v1_20")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_20")))]
 unsafe impl MetaAPI for VideoCodecAlphaMeta {
     type GstType = ffi::GstVideoCodecAlphaMeta;
 
     #[doc(alias = "gst_video_codec_alpha_meta_api_get_type")]
+    #[inline]
     fn meta_api() -> glib::Type {
         unsafe { from_glib(ffi::gst_video_codec_alpha_meta_api_get_type()) }
     }
 }
 
-#[cfg(any(feature = "v1_20", feature = "dox"))]
-#[cfg_attr(feature = "dox", doc(cfg(feature = "v1_20")))]
+#[cfg(feature = "v1_20")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_20")))]
 impl fmt::Debug for VideoCodecAlphaMeta {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         f.debug_struct("VideoCodecAlphaMeta")
@@ -876,6 +906,301 @@ impl fmt::Debug for VideoCodecAlphaMeta {
     }
 }
 
+#[cfg(feature = "v1_22")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_22")))]
+#[repr(transparent)]
+#[doc(alias = "GstVideoSEIUserDataUnregisteredMeta")]
+pub struct VideoSeiUserDataUnregisteredMeta(ffi::GstVideoSEIUserDataUnregisteredMeta);
+
+#[cfg(feature = "v1_22")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_22")))]
+unsafe impl Send for VideoSeiUserDataUnregisteredMeta {}
+#[cfg(feature = "v1_22")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_22")))]
+unsafe impl Sync for VideoSeiUserDataUnregisteredMeta {}
+
+#[cfg(feature = "v1_22")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_22")))]
+impl VideoSeiUserDataUnregisteredMeta {
+    #[doc(alias = "gst_buffer_add_video_sei_user_data_unregistered_meta")]
+    pub fn add<'a>(
+        buffer: &'a mut gst::BufferRef,
+        uuid: &[u8; 16],
+        data: &[u8],
+    ) -> gst::MetaRefMut<'a, Self, gst::meta::Standalone> {
+        skip_assert_initialized!();
+        assert!(!data.is_empty());
+        unsafe {
+            let meta = ffi::gst_buffer_add_video_sei_user_data_unregistered_meta(
+                buffer.as_mut_ptr(),
+                mut_override(uuid.as_ptr()),
+                mut_override(data.as_ptr()),
+                data.len(),
+            );
+
+            Self::from_mut_ptr(buffer, meta)
+        }
+    }
+
+    #[doc(alias = "get_data")]
+    #[inline]
+    pub fn data(&self) -> &[u8] {
+        if self.0.size == 0 {
+            return &[];
+        }
+        // SAFETY: In the C API we have a pointer data and a size variable
+        // indicating the length of the data. Here we convert it to a size,
+        // making sure we read the size specified in the C API.
+        unsafe {
+            use std::slice;
+            slice::from_raw_parts(self.0.data, self.0.size)
+        }
+    }
+
+    #[doc(alias = "get_uuid")]
+    #[inline]
+    pub fn uuid(&self) -> [u8; 16] {
+        self.0.uuid
+    }
+}
+
+#[cfg(feature = "v1_22")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_22")))]
+impl fmt::Debug for VideoSeiUserDataUnregisteredMeta {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.debug_struct("VideoSeiUserDataUnregisteredMeta")
+            .field(
+                "uuid",
+                &format!("0x{:032X}", u128::from_be_bytes(self.uuid())),
+            )
+            .field("data", &self.data())
+            .finish()
+    }
+}
+
+#[cfg(feature = "v1_22")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_22")))]
+unsafe impl MetaAPI for VideoSeiUserDataUnregisteredMeta {
+    type GstType = ffi::GstVideoSEIUserDataUnregisteredMeta;
+
+    #[doc(alias = "gst_video_sei_user_data_unregistered_meta_api_get_type")]
+    fn meta_api() -> glib::Type {
+        unsafe {
+            glib::translate::from_glib(ffi::gst_video_sei_user_data_unregistered_meta_api_get_type())
+        }
+    }
+}
+
+#[cfg(feature = "v1_24")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_24")))]
+#[repr(transparent)]
+#[doc(alias = "GstAncillaryMeta")]
+pub struct AncillaryMeta(ffi::GstAncillaryMeta);
+
+#[cfg(feature = "v1_24")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_24")))]
+unsafe impl Send for AncillaryMeta {}
+#[cfg(feature = "v1_24")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_24")))]
+unsafe impl Sync for AncillaryMeta {}
+
+#[cfg(feature = "v1_24")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_24")))]
+impl AncillaryMeta {
+    #[doc(alias = "gst_buffer_add_ancillary_meta")]
+    pub fn add(buffer: &mut gst::BufferRef) -> gst::MetaRefMut<Self, gst::meta::Standalone> {
+        skip_assert_initialized!();
+        unsafe {
+            let meta = ffi::gst_buffer_add_ancillary_meta(buffer.as_mut_ptr());
+
+            Self::from_mut_ptr(buffer, meta)
+        }
+    }
+
+    #[inline]
+    pub fn field(&self) -> crate::AncillaryMetaField {
+        unsafe { from_glib(self.0.field) }
+    }
+
+    #[inline]
+    pub fn set_field(&mut self, field: crate::AncillaryMetaField) {
+        self.0.field = field.into_glib();
+    }
+
+    #[inline]
+    pub fn c_not_y_channel(&self) -> bool {
+        unsafe { from_glib(self.0.c_not_y_channel) }
+    }
+
+    #[inline]
+    pub fn set_c_not_y_channel(&mut self, c_not_y_channel: bool) {
+        self.0.c_not_y_channel = c_not_y_channel.into_glib();
+    }
+
+    #[inline]
+    pub fn line(&self) -> u16 {
+        self.0.line
+    }
+
+    #[inline]
+    pub fn set_line(&mut self, line: u16) {
+        self.0.line = line;
+    }
+
+    #[inline]
+    pub fn offset(&self) -> u16 {
+        self.0.offset
+    }
+
+    #[inline]
+    pub fn set_offset(&mut self, offset: u16) {
+        self.0.offset = offset;
+    }
+
+    #[inline]
+    pub fn did(&self) -> u16 {
+        self.0.DID
+    }
+
+    #[inline]
+    pub fn set_did(&mut self, did: u16) {
+        self.0.DID = did;
+    }
+
+    #[inline]
+    pub fn sdid_block_number(&self) -> u16 {
+        self.0.SDID_block_number
+    }
+
+    #[inline]
+    pub fn set_sdid_block_number(&mut self, sdid_block_number: u16) {
+        self.0.SDID_block_number = sdid_block_number;
+    }
+
+    #[inline]
+    pub fn data_count(&self) -> u16 {
+        self.0.data_count
+    }
+
+    #[inline]
+    pub fn checksum(&self) -> u16 {
+        self.0.checksum
+    }
+
+    #[inline]
+    pub fn set_checksum(&mut self, checksum: u16) {
+        self.0.checksum = checksum;
+    }
+
+    #[inline]
+    pub fn data(&self) -> &[u16] {
+        if self.0.data_count & 0xff == 0 {
+            return &[];
+        }
+        unsafe {
+            use std::slice;
+
+            slice::from_raw_parts(self.0.data, (self.0.data_count & 0xff) as usize)
+        }
+    }
+
+    #[inline]
+    pub fn data_mut(&mut self) -> &mut [u16] {
+        if self.0.data_count & 0xff == 0 {
+            return &mut [];
+        }
+        unsafe {
+            use std::slice;
+
+            slice::from_raw_parts_mut(self.0.data, (self.0.data_count & 0xff) as usize)
+        }
+    }
+
+    #[inline]
+    pub fn set_data(&mut self, data: glib::Slice<u16>) {
+        unsafe {
+            assert!(data.len() < 256);
+            self.0.data_count = data.len() as u16;
+            self.0.data = data.into_glib_ptr();
+        }
+    }
+
+    #[inline]
+    pub fn set_data_count_upper_two_bits(&mut self, upper_two_bits: u8) {
+        assert!(upper_two_bits & !0x03 == 0);
+        self.0.data_count = ((upper_two_bits as u16) << 8) | self.0.data_count & 0xff;
+    }
+}
+
+#[cfg(feature = "v1_24")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_24")))]
+unsafe impl MetaAPI for AncillaryMeta {
+    type GstType = ffi::GstAncillaryMeta;
+
+    #[doc(alias = "gst_ancillary_meta_api_get_type")]
+    #[inline]
+    fn meta_api() -> glib::Type {
+        unsafe { from_glib(ffi::gst_ancillary_meta_api_get_type()) }
+    }
+}
+
+#[cfg(feature = "v1_24")]
+#[cfg_attr(docsrs, doc(cfg(feature = "v1_24")))]
+impl fmt::Debug for AncillaryMeta {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.debug_struct("AncillaryMeta")
+            .field("field", &self.field())
+            .field("c_not_y_channel", &self.c_not_y_channel())
+            .field("line", &self.line())
+            .field("offset", &self.offset())
+            .field("did", &self.did())
+            .field("sdid_block_number", &self.sdid_block_number())
+            .field("data_count", &self.data_count())
+            .field("data", &self.data())
+            .field("checksum", &self.checksum())
+            .finish()
+    }
+}
+
+pub mod tags {
+    gst::impl_meta_tag!(Video, crate::ffi::GST_META_TAG_VIDEO_STR);
+    gst::impl_meta_tag!(Size, crate::ffi::GST_META_TAG_VIDEO_SIZE_STR);
+    gst::impl_meta_tag!(Orientation, crate::ffi::GST_META_TAG_VIDEO_ORIENTATION_STR);
+    gst::impl_meta_tag!(Colorspace, crate::ffi::GST_META_TAG_VIDEO_COLORSPACE_STR);
+}
+
+#[derive(Debug, Clone, PartialEq, Eq)]
+pub struct VideoMetaTransformScale<'a> {
+    in_info: &'a crate::VideoInfo,
+    out_info: &'a crate::VideoInfo,
+}
+
+impl<'a> VideoMetaTransformScale<'a> {
+    pub fn new(in_info: &'a crate::VideoInfo, out_info: &'a crate::VideoInfo) -> Self {
+        skip_assert_initialized!();
+        VideoMetaTransformScale { in_info, out_info }
+    }
+}
+
+unsafe impl<'a> gst::meta::MetaTransform<'a> for VideoMetaTransformScale<'a> {
+    type GLibType = ffi::GstVideoMetaTransform;
+
+    #[doc(alias = "gst_video_meta_transform_scale_get_quark")]
+    fn quark() -> glib::Quark {
+        unsafe { from_glib(ffi::gst_video_meta_transform_scale_get_quark()) }
+    }
+
+    fn to_raw<T: MetaAPI>(
+        &self,
+        _meta: &gst::MetaRef<T>,
+    ) -> Result<ffi::GstVideoMetaTransform, glib::BoolError> {
+        Ok(ffi::GstVideoMetaTransform {
+            in_info: mut_override(self.in_info.to_glib_none().0),
+            out_info: mut_override(self.out_info.to_glib_none().0),
+        })
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
@@ -902,6 +1227,10 @@ mod tests {
             assert_eq!(meta.n_planes(), 1);
             assert_eq!(meta.offset(), &[0]);
             assert_eq!(meta.stride(), &[320 * 4]);
+            assert!(meta.has_tag::<gst::meta::tags::Memory>());
+            assert!(meta.has_tag::<tags::Video>());
+            assert!(meta.has_tag::<tags::Colorspace>());
+            assert!(meta.has_tag::<tags::Size>());
         }
 
         {
@@ -958,7 +1287,7 @@ mod tests {
 
     #[test]
     #[cfg(feature = "v1_18")]
-    fn test_vide_meta_alignment() {
+    fn test_video_meta_alignment() {
         gst::init().unwrap();
 
         let mut buffer = gst::Buffer::with_size(115200).unwrap();
@@ -1027,4 +1356,73 @@ mod tests {
         assert_eq!(meta.plane_size().unwrap(), [79360, 39680, 0, 0]);
         assert_eq!(meta.plane_height().unwrap(), [248, 124, 0, 0]);
     }
+
+    #[test]
+    #[cfg(feature = "v1_22")]
+    fn test_get_video_sei_user_data_unregistered_meta() {
+        gst::init().unwrap();
+
+        const META_UUID: &[u8; 16] = &[
+            0x4D, 0x49, 0x53, 0x50, 0x6D, 0x69, 0x63, 0x72, 0x6F, 0x73, 0x65, 0x63, 0x74, 0x69,
+            0x6D, 0x65,
+        ];
+
+        const META_DATA: &[u8] = &[
+            0x1f, 0x00, 0x05, 0xff, 0x21, 0x7e, 0xff, 0x29, 0xb5, 0xff, 0xdc, 0x13,
+        ];
+
+        let buffer_data = &[
+            &[0x00, 0x00, 0x00, 0x20, 0x06, 0x05, 0x1c],
+            META_UUID as &[u8],
+            META_DATA,
+            &[
+                0x80, 0x00, 0x00, 0x00, 0x14, 0x65, 0x88, 0x84, 0x00, 0x10, 0xff, 0xfe, 0xf6, 0xf0,
+                0xfe, 0x05, 0x36, 0x56, 0x04, 0x50, 0x96, 0x7b, 0x3f, 0x53, 0xe1,
+            ],
+        ]
+        .concat();
+
+        let mut harness = gst_check::Harness::new("h264parse");
+        harness.set_src_caps_str(r#"
+            video/x-h264, stream-format=(string)avc,
+            width=(int)1920, height=(int)1080, framerate=(fraction)25/1,
+            bit-depth-chroma=(uint)8, parsed=(boolean)true,
+            alignment=(string)au, profile=(string)high, level=(string)4,
+            codec_data=(buffer)01640028ffe1001a67640028acb200f0044fcb080000030008000003019478c1924001000568ebccb22c
+        "#);
+        let buffer = gst::Buffer::from_slice(buffer_data.clone());
+        let buffer = harness.push_and_pull(buffer).unwrap();
+
+        let meta = buffer.meta::<VideoSeiUserDataUnregisteredMeta>().unwrap();
+        assert_eq!(meta.uuid(), *META_UUID);
+        assert_eq!(meta.data(), META_DATA);
+        assert_eq!(meta.data().len(), META_DATA.len());
+    }
+
+    #[test]
+    fn test_meta_video_transform() {
+        gst::init().unwrap();
+
+        let mut buffer = gst::Buffer::with_size(320 * 240 * 4).unwrap();
+        let meta = VideoCropMeta::add(buffer.get_mut().unwrap(), (10, 10, 20, 20));
+
+        let mut buffer2 = gst::Buffer::with_size(640 * 480 * 4).unwrap();
+
+        let in_video_info = crate::VideoInfo::builder(crate::VideoFormat::Rgba, 320, 240)
+            .build()
+            .unwrap();
+        let out_video_info = crate::VideoInfo::builder(crate::VideoFormat::Rgba, 640, 480)
+            .build()
+            .unwrap();
+
+        meta.transform(
+            buffer2.get_mut().unwrap(),
+            &VideoMetaTransformScale::new(&in_video_info, &out_video_info),
+        )
+        .unwrap();
+
+        let meta2 = buffer2.meta::<VideoCropMeta>().unwrap();
+
+        assert_eq!(meta2.rect(), (20, 20, 40, 40));
+    }
 }
